﻿using System.Collections.Generic;
using System.Collections;
using UnityEngine;
using XInputDotNetPure;
using UnityEngine.UI;

public enum WavePhase {
    None = 0,
    Locking,
    Syncing
}

public class WaveController : MonoBehaviour {
    public System.Action onLock;
   
    // How long the player needs to keep the sticks in place after matching
    public float lockTime = 1f;

    // When pointing in the wrong direction, the result will only be half as strong
    public float inverseMultiplier = 0.5f;

    public WaveLine line;
    public PlayerWave player;
    public Text lockedText;
    public GameObject incomingGO;
    public GameObject lockQuad;
    public AudioClip lockingSfx;
    public AudioClip lockedSfx;
    public AudioClip incomingSfx;

    public WavePhase CurrentPhase { get; private set; }

    // The threshold value for when the vertical value is considered correct
    private float verticalThreshold = 0.9f;

    // The threshold value for when the horizontal value is considered correct
    private float horizontalThreshold = 0.90f;

    // How much of the graph is visible by default (without any player input)
    private float defaultAlpha = 0.2f;
    
    private GameObject lineGO;
    private GameObject lockedGO;
    private GameObject playerGO;
    private Transform lockQuadTransform;
    private List<Waves.WaveType> waveTypes = new List<Waves.WaveType>() { new Waves.Sine(), new Waves.CustomSine(), new Waves.Perlin1(), new Waves.Perlin2(), new Waves.Perlin3() };
    private float matchHeight;
    private float matchPeriod;
    private Vector2 randomLeft;
    private Vector2 randomRight;
    private float leftAngle;
    private float rightAngle;
    private float currentLock = 0;
    private float targetAlpha = 1.0f;
    private float animInTime = 0.5f;
    private float inverseDefaultAlpha;
    private Vector3 lockQuadScale;
    private float rightRumbleAmount = 1;
    private float leftRumbleAmount = 0;
    private bool isVibrating = false;
    private AudioController audioController;

    protected void Awake() {
        lineGO = line.gameObject;
        playerGO = player.gameObject;
        lockedGO = lockedText.gameObject;
        lockQuadTransform = lockQuad.transform;
        lockQuadScale = lockQuadTransform.localScale;
        inverseDefaultAlpha = 1 - defaultAlpha;
        SetPhase(WavePhase.None);
    }

    protected void Start() {
        audioController = FindObjectOfType<AudioController>();
    }

    public void GenerateGraph(Waves.WaveTypeLink waveType, float height, float period) {
        randomLeft = new Vector2(Random.Range(-1f, 1f), Random.Range(-1f, 1f)).normalized;
        randomRight = new Vector2(Random.Range(-1, 1f), Random.Range(-1f, 1f)).normalized;

        Waves.WaveType newType = waveTypes[(int)waveType];

        matchHeight = height;
        matchPeriod = period;

        line.SetProperties(height, period);
        player.SetType(newType);
        line.SetType(newType);
    }

    protected void Update() {
        if(CurrentPhase != WavePhase.Locking) {
            if(CurrentPhase == WavePhase.None && Input.GetKeyDown(KeyCode.Space)) {
                SetPhase(WavePhase.Locking);
            }
            return;
        }

        Vector2 left = new Vector2(Input.GetAxis("HorizontalLeft"), Input.GetAxis("VerticalLeft")).normalized;
        Vector2 right = new Vector2(Input.GetAxis("HorizontalRight"), Input.GetAxis("VerticalRight")).normalized;

        leftAngle = Vector2.Dot(randomLeft, left);
        leftAngle = leftAngle < 0 ? leftAngle*0.5f : leftAngle;
        rightAngle = Vector2.Dot(randomRight, right);
        rightAngle = rightAngle < 0 ? rightAngle*0.5f : rightAngle;

        SetAlpha(leftAngle, rightAngle);

        // Decides whether the player is close enough to the actual graph
        bool toggle = leftAngle > horizontalThreshold && rightAngle > verticalThreshold;
        
        player.Toggle(toggle);
        line.Toggle(toggle);

        if(lockQuad.activeInHierarchy != toggle) {
            lockQuad.SetActive(toggle);

            if (lockingSfx != null) {
                if (toggle) {
                    if (!audioController.sfxAudioSource.isPlaying || audioController.sfxAudioSource.clip != lockingSfx) {
                        audioController.PlaySFX(lockingSfx, 0, true);
                    }
                } else if (audioController.sfxAudioSource.clip == lockingSfx) {
                    audioController.sfxAudioSource.Stop();
                }
            }
        }

        if (toggle) {
            currentLock += Time.deltaTime;
            lockQuadTransform.localScale = new Vector3(lockQuadScale.x * Mathf.Clamp01(currentLock / lockTime), lockQuadScale.y, lockQuadScale.z);
            if (currentLock >= lockTime) {
                if (lockedSfx != null) {
                    audioController.PlaySFX(lockedSfx, 0, false);
                }
                SetPhase(WavePhase.Syncing);
                if (onLock != null) {
                    onLock.Invoke();
                }
            }
        } else {
            currentLock = 0;
        }
    }

    public void SetPhase(WavePhase phase) {
        CurrentPhase = phase;
        lockedGO.SetActive(false);
        if (phase == WavePhase.None) {
            lockQuad.SetActive(false);
        }
        incomingGO.SetActive(false);

        if (phase == WavePhase.None) {
            Toggle(false);
            return;
        }
        
        if (phase == WavePhase.Locking) {
            StartCoroutine(IncomingSignal());
        } else if(phase == WavePhase.Syncing) {
            StartCoroutine(SyncGraph());
            lockedGO.SetActive(true);
        }
    }

    public void SetAlpha(float x, float y) {
        // Even if there's no player input from the other axis, make sure the graph changes
        player.SetProperties((defaultAlpha + (x * inverseDefaultAlpha)) * matchHeight, (defaultAlpha + (y * inverseDefaultAlpha)) * matchPeriod);
        float noise = 0.2f +  (x  < 0.9f ? 0.1f : (1 - x)) + (y < 0.9f ? 0.1f : (1 - y));
        player.SetNoise(noise);
        line.SetNoise(noise);
    }

    public void Toggle(bool toggle) {
        playerGO.SetActive(toggle);
        lineGO.SetActive(toggle);
    }

    public void Vibrate(float duration) {
        StartCoroutine(IEVibrate(duration));
    }

    private IEnumerator IEVibrate(float duration) {
        isVibrating = true;
        yield return new WaitForSeconds(duration);
        isVibrating = false;
    }

    private IEnumerator IncomingSignal() {
        incomingGO.SetActive(true);
        
        Vibrate(0.1f);

        if (incomingSfx != null) {
            audioController.PlaySFX(incomingSfx, 0, false);
        } else {
            Debug.Log("Incoming SFX is missing!");
        }
        yield return new WaitForSeconds(1.5f);
        Toggle(true);
        float i = 0;
        float d;
        while (i < animInTime) {
            SetWaveScale(Mathf.Sin((i / animInTime)*Mathf.PI/2));
            d = Time.deltaTime;
            i += d;
            yield return null;
        }

        SetWaveScale(1.0f);
    }

    protected void FixedUpdate() {
        if (CurrentPhase != WavePhase.None) {
            if (isVibrating) {
                GamePad.SetVibration(PlayerIndex.One, leftRumbleAmount, rightRumbleAmount);
            } else {
                GamePad.SetVibration(PlayerIndex.One, 0, 0);
            }
        }
    }

    private void SetWaveScale(float s) {
        player.SetWaveScale(s);
        line.SetWaveScale(s);
    }

    private IEnumerator SyncGraph() {
        float i = 0;
        float d;
        Vibrate(0.1f);
        while (CurrentPhase == WavePhase.Syncing) {
            yield return null;
            d = Time.deltaTime;
            i += d;
            leftAngle = Mathf.Lerp(leftAngle, targetAlpha, d);
            rightAngle = Mathf.Lerp(rightAngle, targetAlpha, d);
            SetAlpha(leftAngle, rightAngle);
        }

        SetAlpha(targetAlpha, targetAlpha);
    }
}
